/**
can.c

Routines for sending and receiving messages for configuration and/or
communication over CAN network using a SJA1000 transceiver.
For use in UL_USB1 module, it runs in Intel mode.
See documentation for details.

*/


#include "can/can.h"

inline void can_data_pins_dir_output(void)
{
	IO1DIR|=P1_SJA1000_DATA_PINS; // Port as output to send data
}

inline void can_data_pins_dir_input(void)
{
	IO1DIR&=~P1_SJA1000_DATA_PINS; // Sets port as input
}

inline void can_data_pins_set_value(uint8_t data)
{
	uint32_t val = __val2mfld(P1_SJA1000_DATA_PINS,data);
	/*
	 * Clear only that pins, which need that, lower transition
	 * frequency and eliminate spikes
	 */
	IO1CLR= val ^ P1_SJA1000_DATA_PINS;
	IO1SET= val;
}

inline uint8_t can_data_pins_get_value(void)
{
	return __mfld2val(P1_SJA1000_DATA_PINS,IO1PIN);
}


void can_comm_init()
{
	//CANMSG("Start can_comm_init\n");
	IO1DIR |= P1_OUT_PORT_CS_PIN|P1_SJA1000_RST_PIN;
	// Due to change in design there is CS_PIN connected with ALE_PIN and ALE_PIN connection to LPC is interrupted
	// We don't use ALE_PIN
	//IO0DIR|=P0_SJA1000_ALE_PIN|P0_SJA1000_CS_PIN|P0_SJA1000_RD_PIN|P0_SJA1000_WR_PIN;
	IO0DIR|=P0_SJA1000_CS_PIN|P0_SJA1000_RD_PIN|P0_SJA1000_WR_PIN;
	IO0DIR&=~(P0_SJA1000_INT_PIN);

	SET_OUT_PIN(IO0,P0_SJA1000_CS_PIN);
	CLR_OUT_PIN(IO1,P1_SJA1000_RST_PIN);
	SJA1000_INIT_DELAY();
	SET_OUT_PIN(IO0,P0_SJA1000_CS_PIN);
	// Due to change in design there is CS_PIN connected with ALE_PIN and ALE_PIN connection to LPC is interrupted
	//	CLR_OUT_PIN(IO0,P0_SJA1000_ALE_PIN);
	SET_OUT_PIN(IO1,P1_SJA1000_RST_PIN);
	SJA1000_INIT_DELAY();
}

void can_write(uint8_t data, uint8_t address)
{
	can_data_pins_dir_output();
	// Set memory address
	can_data_pins_set_value(address);
	// Init
	SET_OUT_PIN(IO0,P0_SJA1000_RD_PIN);	// Stays high on write
	SET_OUT_PIN(IO0,P0_SJA1000_WR_PIN); // Stays high on address write
	SET_OUT_PIN(IO0,P0_SJA1000_CS_PIN); // Sets output buffers to third state
	//SJA1000_DELAY();
	//SET_OUT_PIN(IO0,P0_SJA1000_ALE_PIN); // Start command

	//SJA1000_DELAY();
	//CLR_OUT_PIN(IO0,P0_SJA1000_ALE_PIN); // Makes address active
	CLR_OUT_PIN(IO0,P0_SJA1000_CS_PIN);
	//SJA1000_DELAY();

	// Setting data
	can_data_pins_set_value(data);

	CLR_OUT_PIN(IO0,P0_SJA1000_WR_PIN);
	CLR_OUT_PIN(IO0,P0_SJA1000_WR_PIN);
	//SJA1000_DELAY();
	SET_OUT_PIN(IO0,P0_SJA1000_WR_PIN); // Data should be accepted by now
	SET_OUT_PIN(IO0,P0_SJA1000_CS_PIN);
	//SJA1000_DELAY();
}

uint8_t can_read(const uint8_t address)
{
	uint8_t data;

	can_data_pins_dir_output();
	// Set memory address
	can_data_pins_set_value(address);
	// Init
	SET_OUT_PIN(IO0,P0_SJA1000_WR_PIN); // Stays high on read
	SET_OUT_PIN(IO0,P0_SJA1000_RD_PIN); // Stays high while entering address
	SET_OUT_PIN(IO0,P0_SJA1000_CS_PIN);
	//SJA1000_DELAY();
	//SET_OUT_PIN(IO0,P0_SJA1000_ALE_PIN);

	//SJA1000_DELAY();
	//CLR_OUT_PIN(IO0,P0_SJA1000_ALE_PIN);
	CLR_OUT_PIN(IO0,P0_SJA1000_CS_PIN);

	// Get data

	can_data_pins_dir_input();
	CLR_OUT_PIN(IO0,P0_SJA1000_RD_PIN);
	CLR_OUT_PIN(IO0,P0_SJA1000_RD_PIN);
	//SJA1000_DELAY();
	data = can_data_pins_get_value();
	SET_OUT_PIN(IO0,P0_SJA1000_RD_PIN);
	SET_OUT_PIN(IO0,P0_SJA1000_CS_PIN);
	//SJA1000_DELAY();

	return data;
}

int can_init(){
	uint8_t data=0,count=0;
	do {
 		can_comm_init();
		data = can_read(SJAMOD);
		if (count++ > 50)
			return -1;
	} while (!(data&sjaMOD_RM));

	data=sjaCDR_CLKOUT_DIV1|sjaCDR_CLK_OFF|sjaCDR_CBP|sjaCDR_PELICAN;
	can_write(data, SJACDR);

	// Single acceptance filter, reset mode
	data=sjaMOD_AFM|sjaMOD_RM;
	can_write(data, SJAMOD);

	// Enabling all interrupt sources
	data=sjaENABLE_INTERRUPTS;
	can_write(data, SJAIER);

	// Accept all messages
	data=0xFF;
	can_write(data, SJAAMR0);
	can_write(data, SJAAMR0+1);
	can_write(data, SJAAMR0+2);
	can_write(data, SJAAMR0+3);

	data=sjaCMR_CDO;
	can_write(data, SJACMR);

	return 0;
}
