/**************************************************************************/
/* File: finish.c - finalization of the driver operations                 */
/*                                                                        */
/* LinCAN - (Not only) Linux CAN bus driver                               */
/* Copyright (C) 2002-2009 DCE FEE CTU Prague <http://dce.felk.cvut.cz>   */
/* Copyright (C) 2002-2009 Pavel Pisa <pisa@cmp.felk.cvut.cz>             */
/* Funded by OCERA and FRESCOR IST projects                               */
/* Based on CAN driver code by Arnaud Westenberg <arnaud@wanadoo.nl>      */
/*                                                                        */
/* LinCAN is free software; you can redistribute it and/or modify it      */
/* under terms of the GNU General Public License as published by the      */
/* Free Software Foundation; either version 2, or (at your option) any    */
/* later version.  LinCAN is distributed in the hope that it will be      */
/* useful, but WITHOUT ANY WARRANTY; without even the implied warranty    */
/* of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU    */
/* General Public License for more details. You should have received a    */
/* copy of the GNU General Public License along with LinCAN; see file     */
/* COPYING. If not, write to the Free Software Foundation, 675 Mass Ave,  */
/* Cambridge, MA 02139, USA.                                              */
/*                                                                        */
/* To allow use of LinCAN in the compact embedded systems firmware        */
/* and RT-executives (RTEMS for example), main authors agree with next    */
/* special exception:                                                     */
/*                                                                        */
/* Including LinCAN header files in a file, instantiating LinCAN generics */
/* or templates, or linking other files with LinCAN objects to produce    */
/* an application image/executable, does not by itself cause the          */
/* resulting application image/executable to be covered by                */
/* the GNU General Public License.                                        */
/* This exception does not however invalidate any other reasons           */
/* why the executable file might be covered by the GNU Public License.    */
/* Publication of enhanced or derived LinCAN files is required although.  */
/**************************************************************************/

#include "../include/can.h"
#include "../include/can_sysdep.h"
#include "../include/main.h"
#include "../include/devcommon.h"
#include "../include/finish.h"
#include "../include/setup.h"

extern int next_minor;
/**
 * msgobj_done - destroys one CAN message object
 * @obj: pointer to CAN message object structure
 */
void msgobj_done(struct msgobj_t *obj)
{
	int delayed=0;
	if(obj->qends) {
		delayed=canqueue_ends_done_chip(obj->qends);
		if(delayed < 0)
			CANMSG("msgobj_done: problem with chip queue ends\n");
	}

	if((obj->hostchip) && (obj->object>0)) {
		if(obj->hostchip->msgobj[obj->object-1] == obj)
			obj->hostchip->msgobj[obj->object-1]=NULL;
		else
			CANMSG("msgobj_done: not registered in the canchip_t\n");
		obj->hostchip=NULL;
	}

	if((obj->minor>=0)) {
		if(objects_p[obj->minor] == obj){
			objects_p[obj->minor] = NULL;
			if (--next_minor<0)
				next_minor=0;
		}
		else
			CANMSG("msgobj_done: not registered as minor\n");
	}

	del_timer_sync(&obj->tx_timeout);

	if(obj->qends) {
		/*delayed free could be required there in the future,
		  actual use patter cannot generate such situation*/
		if(!delayed) {
			can_checked_free(obj->qends);
		}
	}
	obj->qends=NULL;
}


/**
 * canchip_done - destroys one CAN chip representation
 * @chip: pointer to CAN chip structure
 */
void canchip_done(struct canchip_t *chip)
{

	int i;
	struct msgobj_t *obj;

	if(chip->flags & CHIP_ATTACHED){
		chip->chipspecops->release_chip(chip);
		chip->flags &= ~CHIP_ATTACHED;
	}

	if((chip->hostdevice) && (chip->chip_idx>=0)) {
		if(chip->hostdevice->chip[chip->chip_idx] == chip)
			chip->hostdevice->chip[chip->chip_idx] = NULL;
		else
			CANMSG("canchip_done: not registered in hostdevice\n");
	}

	can_chip_free_irq(chip);

	can_synchronize_irq(chip->chip_irq);

  	for(i=0; i<chip->max_objects; i++){
		if((obj=chip->msgobj[i])==NULL)
			continue;
		msgobj_done(obj);
		can_checked_free(obj);
	}

	can_checked_free(chip->chipspecops);
       
	if(~chip->flags & CHIP_KEEP_DATA)
		can_checked_free(chip->chip_data);
	chip->chip_data = NULL;
	chip->chipspecops=NULL;

}

/**
 * candevice_done - destroys representation of one CAN device/board
 * @candev: pointer to CAN device/board structure
 */
void candevice_done(struct candevice_t *candev)
{
	int i;
	struct canchip_t *chip;

  	for(i=0; i<candev->nr_all_chips; i++){
		if((chip=candev->chip[i])==NULL)
			continue;
		canchip_done(chip);
		can_checked_free(chip);

	}
	if(candev->flags & CANDEV_IO_RESERVED) {
		candev->hwspecops->release_io(candev);
		candev->flags &= ~CANDEV_IO_RESERVED;
	}
	can_checked_free(candev->hwspecops);
	candev->hwspecops=NULL;
}

/**
 * candevice_done - destroys representation of all CAN devices/boards
 * @canhw: pointer to the root of all CAN hardware representation
 */
void canhardware_done(struct canhardware_t *canhw)
{
	int i;
	struct candevice_t *candev;

  	for(i=0; i<MAX_HW_CARDS; i++){
		if((candev=canhw->candevice[i])==NULL)
			continue;
		candevice_done(candev);
		can_checked_free(candev);
	}

}
